__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.main import Client
from contributor_ci.logger import logger


def main(args, parser, extra, subparser):

    client = Client(
        quiet=args.quiet,
        config_file=args.config_file,
        outdir=args.outdir,
        require_config=False,
    )

    if not args.username.startswith("user:") and not args.username.startswith("org:"):
        logger.exit("You must provide a user:<username> or org:<orgname>")
    client.init(args.username)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.main.extractor import GitHubExtractorBase
from contributor_ci.logger import logger


class RepoMetadata(GitHubExtractorBase):

    name = "repo_metadata"
    description = "gather repository metadata from several extractors."
    depends_on = ["repos", "topics"]
    filenames = ["repo_metadata"]

    def extract(self):
        """
        Gather repository metadata from multiple extractors
        """
        self._data[self.name] = {}

        # Load in dependency files
        repos = self.load_dependency_file("repos")
        topics = self.load_dependency_file("topics")

        for repo in repos.data:
            logger.info("Combining metadata for repository %s" % repo)
            data = {}

            repository = repos.data[repo]
            data["name"] = repo
            data["description"] = repository["description"]
            data["website"] = repository["homepageUrl"]

            # Add topis
            if repository["repositoryTopics"]["totalCount"] > 0 and repo in topics.data:
                topic_repo = topics.data[repo]
                topiclist = []
                for topic in topic_repo["repositoryTopics"]["nodes"]:
                    topiclist.append(topic["topic"]["name"])
                data["topics"] = topiclist
            else:
                data["topics"] = None

            self._data[self.name][repo] = data



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"


from contributor_ci.logger import logger
from contributor_ci.cfa import CFA
from contributor_ci.main.extractors.users import UserExtractor

import contributor_ci.utils as utils
from .settings import Settings, SettingsBase
from .extractor import ExtractorFinder, ExtractorResolver

import os
import shutil
import sys
import random


class Client:
    """
    Create a client to extract metrics and run Contributor CI.
    """

    def __init__(self, config_file=None, outdir=None, quiet=True, require_config=True):
        self.quiet = quiet
        self._outdir = outdir
        self._extractors = None
        self._results = {}
        self._init_settings(config_file, require_config)

    def _init_settings(self, config_file, require_config=True):
        """
        Init settings.
        """
        if require_config:
            self.settings = Settings(config_file)
        else:
            self.settings = SettingsBase()

    @property
    def out_dir(self):
        """
        Get the output directory for contributor ci results
        """
        if not hasattr(self, "_out_dir"):
            default = os.path.abspath(".cci")
            self._out_dir = self._outdir or self.settings.outdir or default
        return self._out_dir

    @property
    def extractors(self):
        """
        Get a list of extractirs available
        """
        if not self._extractors:
            self._finder = ExtractorFinder()
            self._extractors = dict(self._finder.items())
        return self._extractors

    def init(self, username):
        """
        Create a new contributor-ci.yaml based on an org or username.
        """
        if not self.settings.config_file:
            self.settings.config_file = "contributor-ci.yaml"

        # If the config file specified already exists, no go
        if os.path.exists(self.settings.config_file):
            logger.exit(
                "%s already exists, remove to overwrite." % self.settings.config_file
            )
        result = UserExtractor().extract(username)
        # Validate and save the new config
        self.settings._settings = result
        self.settings.validate()
        self.settings.save(self.settings.config_file)

    def ui_generate(self, dirname, include_cfa=False):
        """
        Generate a new user interface.
        """
        dirname = os.path.abspath(dirname)

        # We've already checked for a config file!
        # If we already have a site here, no go
        for filename in ["_config.yml", "_layouts"]:
            if os.path.exists(os.path.join(dirname, filename)):
                logger.exit("A CCI site already exists here!")

        # Copy the entire contents of the site folder here
        site = os.path.join(utils.get_installdir(), "site")
        utils.copytree(site, dirname)
        self.ui_update(dirname, include_cfa=include_cfa)

    def ui_update(self, dirname, options=None, include_cfa=False):
        """
        Update an existing user interface.
        """
        options = options or {}
        dirname = os.path.abspath(dirname)

        # We've already checked for a config file!
        # If we already have a site here, no go
        for filename in ["_config.yml", "_layouts"]:
            if not os.path.exists(os.path.join(dirname, filename)):
                logger.exit("Missing CCI site. Run cci ui generate.")

        # Data is expected to be in cci
        self._out_dir = os.path.join(dirname, "cci")

        # Do we want to extract a random number?
        if "random" in options:
            self.extract_random(options["random"])
        else:
            self.extract_all()

        # CFA files will be here if they exist
        outdir = os.path.join(dirname, "_cfa")

        # Next, are we including CFA?
        if not os.path.exists(outdir):
            logger.info("%s has been updated." % dirname)
            return

        # Provide a different data directory
        if include_cfa:
            cfa = CFA(data_dir=self._out_dir)
            cfa.run_all(outdir)

    def cfa(self, repo, save=False):
        """
        Run the contributor friendliness assessment.

        A CFA assessment comes down to assessing a repository programatically
        for attributes of "contributor friendliness," either updating an
        existing markdown file with metadata or generating a new one if
        it does not exist.
        """
        cfa = CFA()

        # Save all to the cfa output directory, if don't exist
        outdir = os.path.join(self.out_dir, "cfa")
        if repo == "all" or save and not os.path.exists(outdir):
            os.makedirs(outdir)

        # If cfa is for all repos, we require a config
        if repo == "all":
            cfa.run_all(outdir)

        elif save:
            logger.info("Saving %s to %s" % (repo, outdir))
            return cfa.run(repo, save_to=outdir)

        else:
            return cfa.run(repo)

    def extract(self, method):
        """
        Extract metrics given a particular method
        """
        if method not in self.extractors:
            logger.exit("Extractor %s is not known." % method)

        ext = self.get_extractor(method)

        # Each extractor defines depends_on, and we need to create a DAG that
        # checks dependencies first (either running or finding that already run)
        lookup = {}
        resolver = ExtractorResolver()

        # Keep going until no more dependencies
        expanded = set([ext])
        seen = set()
        while expanded:
            ext = expanded.pop()
            lookup[ext.name] = ext
            if ext in seen:
                continue
            seen.add(ext)
            resolver.add_extractor(ext.name)
            for depname in ext.depends_on:
                dep = self.get_extractor(depname)
                resolver.add_extractor(dep.name)
                resolver.add_dependency(ext.name, dep.name)
                expanded.add(dep)

        order = resolver.resolve()

        ran = False
        for name in order:
            extractor = lookup[name]

            # Only run the extractor if the result does not exist
            if not extractor.exists():
                logger.info("## RUNNING %s extractor" % name)
                self._results[extractor.name] = extractor.extract()
                extractor.save_json()
                ran = True

        # No extractors to run?
        if not ran:
            logger.info("Extractor runs are up to date.")

    def extract_random(self, number):
        """
        Extract a random selection.
        """
        names = list(self.extractors.keys())
        for _ in range(number):

            # Cut out early if we run out!
            if not names:
                return
            choice = random.choice(names)
            names.pop(names.index(choice))
            self.extract(choice)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[contributor-ci-client]"

    def extract_all(self, versions=None):
        """
        Run all extractors
        """
        for name in self.extractors:
            self.extract(name)

    def get_extractor(self, name):
        """
        Load the extractor class based on its name and add settings.
        """
        ext = self._extractors[name]
        ext.settings = self.settings
        ext.outdir = self.out_dir
        return ext



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.logger import logger


def main(args, parser, extra, subparser):

    from contributor_ci.main import Client

    cli = Client(quiet=args.quiet, config_file=args.config_file, outdir=args.outdir)

    # Allowed extractors
    available = list(cli.extractors) + ["all"]
    if args.method not in available:
        logger.exit(
            "%s is not a valid extraction method! Choose from %s." % ",".join(available)
        )

    if args.method == "all":
        cli.extract_all()
    else:
        cli.extract(args.method)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"


from contributor_ci.logger import logger
from contributor_ci.main.extractors.repos import RepoExtractor, ReposExtractor
import contributor_ci.utils as utils
from jinja2 import Template

from copy import deepcopy
import os
import shutil
import sys

here = os.path.abspath(os.path.dirname(__file__))


class CFA:
    """
    Create a client to assess contributor friendliness
    """

    def __init__(self, template=None, data_dir=None):
        self.config = utils.read_yaml(os.path.join(here, "cfa.yaml"))
        self.template = template or os.path.join(here, "template.md")
        self.data_dir = data_dir
        self._cache = {}

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[contributor-friendliness-assesser]"

    def run(self, repo, save_to=None):
        """
        Run an entire contributor friendliness assessment for a repository
        """
        git = utils.GitManager()

        # Determine if we have a path or repository
        if os.path.exists(os.path.abspath(repo)):
            git.update_repo(os.path.abspath(repo))

        # Assume we have git repository to clone
        elif repo.startswith("http"):
            git.clone(repo, dest=utils.get_tmpdir())

        # Cut out early if the file exists
        outfile = os.path.join(save_to, "cfa-%s.md" % git.reponame_flat)
        if os.path.exists(outfile):
            return utils.read_file(outfile)

        # Generate results
        logger.info("Generating CFA for %s" % repo)
        result = self.extract_criteria(git)
        rendered = Template(utils.read_file(self.template)).render(
            items=result, repository=git.reponame
        )

        # If we want to save to file
        if save_to:
            if not os.path.exists(save_to):
                os.makedirs(save_to)
            utils.write_file(outfile, rendered)

        # Clean up
        if os.path.exists(git.folder):
            shutil.rmtree(git.folder)

        return rendered

    def run_all(self, save_to):
        """
        Run a contributor friendliness assessment for all repos known.
        
        Save to (outdir) is required here.
        """
        repos = self.get_cached_result("repos", ReposExtractor)
        for repo in repos:
            self.run("https://github.com/%s" % repo, save_to)

    def get_cached_result(self, name, Extractor, key=None):
        """
        Get a cached result - useful for metrics that use common data.
        """
        # Return cached result if already exists
        if name in self._cache:
            return self._cache[name]

        extractor = Extractor()
        if self.data_dir:
            extractor.outdir = self.data_dir

        # Otherwise, generate the data
        if key:
            meta = extractor.extract(key, reuse=True)
        else:
            meta = extractor.extract(reuse=True)

        # Save new result to cache and return
        self._cache[name] = meta
        return meta

    def extract_criteria(self, repo):
        """
        Given a repository, run methods to extract criteria details
        """
        result = deepcopy(self.config)

        # Run any functions defined in the config
        for key, content in result.items():
            for crit in content.get("criteria", {}):
                if "method" in crit:
                    func = getattr(self, crit["method"], None)
                    if func:
                        metric = func(repo)
                        if metric:
                            crit["result"] = metric
                        if metric and crit.get("type") == "boolean":
                            crit["met"] = True
                    else:
                        logger.warning("CFA does not have method %s" % crit["method"])

        return result

    def count_stars(self, repo):
        meta = self.get_cached_result("repo", RepoExtractor, repo.reponame)
        return meta["stargazers"]["totalCount"]

    def count_forks(self, repo):
        meta = self.get_cached_result("repo", RepoExtractor, repo.reponame)
        return meta["forks"]["totalCount"]

    def get_description(self, repo):
        meta = self.get_cached_result("repo", RepoExtractor, repo.reponame)
        return meta.get("description")

    def has_license(self, repo):
        meta = self.get_cached_result("repo", RepoExtractor, repo.reponame)
        license = meta.get("licenseInfo", {})
        if license:
            return license.get("spdxId") is not None

    def get_license_type(self, repo):
        meta = self.get_cached_result("repo", RepoExtractor, repo.reponame)
        license = meta.get("licenseInfo", {})
        if license:
            return license.get("spdxId")

    def get_build_framework(self, repo):
        meta = self.get_cached_result("repo", RepoExtractor, repo.reponame)
        pass

    def get_url(self, repo):
        meta = self.get_cached_result("repo", RepoExtractor, repo.reponame)
        return meta.get("homepageUrl")



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.main import Client
import re


def main(args, parser, extra, subparser):

    client = Client(
        quiet=args.quiet,
        config_file=args.config_file,
        outdir=args.outdir,
        require_config=False,
    )

    for name, extractor in client.extractors.items():
        if not args.query:
            print("%20s: %s" % (name, extractor.description))
        elif args.query and re.search(args.query, name):
            print("%20s: %s" % (name, extractor.description))



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.logger import logger


def main(args, parser, extra, subparser):

    from contributor_ci.main import Client

    # We require a config to generate for all
    require_config = False
    if args.repo == "all":
        require_config = True

    cli = Client(
        quiet=args.quiet,
        config_file=args.config_file,
        outdir=args.outdir,
        require_config=require_config,
    )
    markdown = cli.cfa(args.repo, save=not args.terminal)
    if args.terminal and args.repo == "all":
        logger.warning("Terminal print is not supported for multiple CFA assessments.")
    elif args.terminal:
        print(markdown)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.logger import logger
from contributor_ci.main.extractor import GitHubExtractorBase

import os


class RepoExtractor(GitHubExtractorBase):
    """
    A separate extractor used by the CFA to get metadata for one repository.
    """

    def extract(self, repo=None, reuse=False):

        repo_owner, repo_name = repo.split("/")

        # Can we re-use an older result?
        if reuse:
            repos = self.load_recent_dependency_file("repos")
            if repos is not None and repo in repos.data:
                return repos.data[repo]

        repo_query = self.get_local_query("repos-info.gql")
        logger.info("\nRetrieving repository info for %s" % repo)
        try:
            out = self.manager.queryGitHubFromFile(
                repo_query, {"ownName": repo_owner, "repoName": repo_name}
            )
        except Exception as error:
            logger.exit("Warning: Could not complete '%s'\n%s" % (repo, error))

        return out["data"]["repository"]


class Repos(GitHubExtractorBase):

    name = "repos"
    description = "extract repository metrics."
    filenames = ["repos"]

    def extract(self):
        """
        Extract repository metadata for orgs and repos.
        """
        self.extract_orgs()
        self.extract_repos()

    def extract_orgs(self):
        """
        Extract list of orgs from the contributor-ci.yaml and update data.
        """
        org_query = self.get_local_query("org-repos-info.gql")

        # This extract only saves one result - a repos file
        self._data[self.name] = {}

        # Exclude these repos
        excludes = self.settings.exclude_repos if self.settings.exclude_repos else []

        # This is going to retrieve repo-level data across the orgs
        for org in self.settings.orgs:
            logger.info("\nRetrieving organization info for %s" % org)

            try:
                out = self.manager.queryGitHubFromFile(
                    org_query,
                    {"orgName": org, "numRepos": 50, "pgCursor": None},
                    paginate=True,
                    cursorVar="pgCursor",
                    keysToList=["data", "organization", "repositories", "nodes"],
                )
            except Exception as error:
                logger.warning("Warning: Could not complete '%s'\n%s" % (org, error))
                continue

            # Add repositories from organization to our data
            for repo in out["data"]["organization"]["repositories"]["nodes"]:

                # Do not add repos that are explicitly excluded!
                if repo["nameWithOwner"] not in excludes:
                    self._data[self.name][repo["nameWithOwner"]] = repo

    def extract_repos(self):
        """
        Add metadata for additional isolated repos in the contibutor-ci.yaml
        """
        repo_query = self.get_local_query("repos-info.gql")

        # Add additional individual repos
        for repo in self.settings.repos:

            # Did we already extract it via an organization?
            if repo in self._data[self.name] or (
                self.settings.exclude_repos and repo in self.settings.exclude_repos
            ):
                continue
            logger.info("\nRetrieving repository info for %s" % repo)
            repo_owner, repo_name = repo.split("/")
            try:
                out = self.manager.queryGitHubFromFile(
                    repo_query, {"ownName": repo_owner, "repoName": repo_name}
                )
            except Exception as error:
                logger.warning("Warning: Could not complete '%s'\n%s" % (repo, error))
                continue

            key = out["data"]["repository"]["nameWithOwner"]
            self._data[self.name][key] = out["data"]["repository"]


class ReposExtractor(Repos):
    """
    A separate extractor used by the CFA to get metadata for all repos
    """

    def extract(self, reuse=False):

        # Can we re-use an older result?
        if reuse:
            repos = self.load_recent_dependency_file("repos")
            if repos is not None:
                return repos.data

        self.extract_orgs()
        self.extract_repos()
        return self._data



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"


## ContributorCI Config Schema

schema_url = "https://json-schema.org/draft-07/schema/#"

unique_array = {"type": "array", "uniqueItems": True, "items": {"type": "string"}}
unique_array_or_null = {
    "oneOf": [
        {"type": "array", "uniqueItems": True, "items": {"type": "string"}},
        {"type": "null"},
    ]
}

configProperties = {
    "member_orgs": unique_array_or_null,
    "orgs": unique_array,
    "repos": unique_array,
    "exclude_repos": unique_array,
    "outdir": {"type": ["string", "null"]},
    "editor": {"type": ["string", "null"]},
}

config_schema = {
    "$schema": schema_url,
    "title": "Contributor CI Config Schema",
    "type": "object",
    "required": ["orgs", "member_orgs",],
    "additionalProperties": False,
    "properties": configProperties,
}
